from oscar.core.loading import get_model

from ocyan.plugin.oscar_communications.customer_extension import utils as ocyan

CommunicationEvent = get_model("order", "CommunicationEvent")


# pylint: disable=arguments-differ,unused-argument,protected-access,arguments-renamed
class Dispatcher(ocyan.Dispatcher):
    def send_email_messages(
        self, recipient_email, messages, from_email=None, attachments=None
    ):
        assert attachments is None
        return super().send_email_messages(recipient_email, messages)

    def create_communication_event(self, order, event_type, dispatched_messages):
        """
        Create order communications event for audit
        """
        if dispatched_messages and event_type is not None:
            CommunicationEvent._default_manager.create(
                order=order, event_type=event_type
            )


class BccDispatcher(Dispatcher):
    def __init__(self, bcc_admins=False, logger=None, mail_connection=None):
        super(BccDispatcher, self).__init__(
            logger=logger, mail_connection=mail_connection
        )
        self.bcc_admins = bcc_admins



import sequences
from django.core.files.base import ContentFile
from django.db import transaction
from weasyprint import CSS, HTML

from oscar.core.loading import get_model

import oscar_invoices.utils

from ocyan.core.fender import config

LegalEntity = get_model("oscar_invoices", "LegalEntity")


class InvoiceCreator(oscar_invoices.utils.InvoiceCreator):
    def __init__(self, **context):
        self.context = context

    def get_legal_entity(self, **kwargs):
        order = kwargs.get("order")
        if order is not None:
            legal_entity = LegalEntity.objects.filter(shop_name=order.site.name).first()

            if legal_entity:
                return legal_entity

        return super().get_legal_entity()

    def create_invoice_model(self, **kwargs):
        kwargs["legal_entity"] = self.get_legal_entity(order=kwargs.get("order"))
        return super().create_invoice_model(**kwargs)

    def generate_invoice_number(self, order, **kwargs):
        return "%06d" % sequences.get_next_value("oscar_invoices")

    def generate_document(self, invoice, **kwargs):
        """
        Create and save invoice document (as *.html file).
        """

        # adds use_path so logo will be rendered as file:// locally accesible
        html = HTML(
            string=self.render_document(
                invoice, use_path=True, **self.context, **kwargs
            )
        )
        css = CSS(string="@page { size: A4; margin: 1cm }")
        content = html.write_pdf(stylesheets=[css])
        return ContentFile(content)

    def get_invoice_filename(self, invoice):
        invoice_file_name = config.get(
            "oscar_communications", "invoice_file_name", "invoice_"
        )
        return "%s%s.pdf" % (invoice_file_name, invoice.order.number)

    def get_or_create_invoice(self, order, **extra_kwargs):
        with transaction.atomic():
            Invoice = self.get_invoice_model()
            try:
                return Invoice.objects.get(order=order)
            except Invoice.DoesNotExist:
                return self.create_invoice(order, **extra_kwargs)



# pylint: disable=abstract-method
import mimetypes
from os.path import basename

from django.conf import settings
from django.core.mail import EmailMessage, EmailMultiAlternatives

from ocyan.main.loading import get_baseclass
from ocyan.plugin.oscar.info import OSCAR_NEEDS_OXYAN

if OSCAR_NEEDS_OXYAN:
    OscarDispatcher = get_baseclass(
        "communication.utils",
        "Dispatcher",
        "ocyan.plugin.oscar_communications.customer_extension",
    )
else:
    OscarDispatcher = get_baseclass(
        "customer.utils",
        "Dispatcher",
        "ocyan.plugin.oscar_communications.customer_extension",
    )


class Dispatcher(OscarDispatcher):
    def __init__(self, logger=None, mail_connection=None):
        super().__init__(logger=logger, mail_connection=mail_connection)
        self.bcc_admins = False

    def send_email_messages(self, recipient, messages):
        """
        Send email to recipient, HTML attachment optional.
        """
        if hasattr(settings, "OSCAR_FROM_EMAIL"):
            from_email = settings.OSCAR_FROM_EMAIL
        else:
            from_email = None

        # Determine whether we are sending a HTML version too
        if messages["html"]:
            email = EmailMultiAlternatives(
                messages["subject"],
                messages["body"],
                from_email=from_email,
                to=[recipient],
            )
            email.attach_alternative(messages["html"], "text/html")
        else:
            email = EmailMessage(
                messages["subject"],
                messages["body"],
                from_email=from_email,
                to=[recipient],
            )
        if "attachments" in messages:
            for attachment in messages["attachments"]:
                content_type, _ = mimetypes.guess_type(attachment.name)
                email.attach(basename(attachment.name), attachment.read(), content_type)

        if self.bcc_admins:
            email.bcc = list(settings.DEFAULT_TO_EMAIL)

        self.logger.info("Sending email to %s", email.recipients())

        if self.mail_connection:
            self.mail_connection.send_messages([email])
        else:
            email.send()

        return email

    def dispatch_order_messages(self, order, messages, event_type=None, **kwargs):
        """
        Dispatch order-related messages to the customer.
        """
        has_billing_address = order.billing_address and order.billing_address.email

        if order.is_anonymous:
            email = kwargs.get("email_address", order.guest_email)
            if has_billing_address and email != order.billing_address.email:
                messages = self.send_billing_email(order, messages)
            dispatched_messages = self.dispatch_anonymous_messages(email, messages)
        else:
            if has_billing_address and order.user.email != order.billing_address.email:
                messages = self.send_billing_email(order, messages)
            dispatched_messages = self.dispatch_user_messages(order.user, messages)

        self.create_communication_event(order, event_type, dispatched_messages)

    def send_billing_email(self, order, messages):
        self.dispatch_direct_messages(order.billing_address.email, messages)
        # We pop out the attachments if there is a billing email because we don't want to send the invoice to the user email.
        messages.pop("attachments", None)
        return messages


class BccDispatcher(Dispatcher):
    def __init__(self, bcc_admins=False, logger=None, mail_connection=None):
        super(BccDispatcher, self).__init__(
            logger=logger, mail_connection=mail_connection
        )
        self.bcc_admins = bcc_admins



from django.contrib import admin

from oscar.core.loading import get_model

from oscar_invoices.admin import InvoiceAdmin as BaseInvoiceAdmin

Invoice = get_model("oscar_invoices", "Invoice")


class InvoiceAdmin(BaseInvoiceAdmin):
    search_fields = (
        "number",
        "order__number",
        "order__user__email",
        "order__guest_email",
    )
    date_hierarchy = "date"
    list_display = ["number", "order", "email"]

    def email(self, invoice):
        if invoice.order:
            return invoice.order.email
        return None


admin.site.unregister(Invoice)
admin.site.register(Invoice, InvoiceAdmin)



from django.utils.translation import gettext_lazy as _

from oscar.apps.customer.alerts.utils import (
    AlertsDispatcher as BaseAlertsDispatcher,
)

from ocyan.main.loading import get_ocyan_class

get_default_site = get_ocyan_class("wagtail_utils.utils", "get_default_site")
get_offline_request = get_ocyan_class("wagtail_utils.utils", "get_offline_request")


class AlertsDispatcher(BaseAlertsDispatcher):
    def get_extra_context(self, alert):
        context = super().get_extra_context(alert)
        site = get_default_site()
        context["request"] = get_offline_request(site.domain)
        return context



from django.db import transaction
from django.db.utils import IntegrityError
from django.test import TestCase

from oscar.core.loading import get_model

from ocyan.main.loading import get_ocyan_class

from ocyan.plugin.oscar_communications.oscar_invoices_extension.utils import (
    InvoiceCreator,
)

Order = get_model("order", "Order")

get_offline_request = get_ocyan_class("wagtail_utils.utils", "get_offline_request")


class InvoiceCreatorTest(TestCase):
    fixtures = [
        "communications/auth",
        "communications/address",
        "communications/catalogue",
        "communications/order",
        "communications/basket",
        "communications/customer",
        "communications/invoices",
    ]

    def check_sequence(self, invoice_creator):
        Invoice = invoice_creator.get_invoice_model()
        last_number = None
        for invoice in Invoice.objects.all():
            if last_number is not None:
                self.assertEqual(int(last_number) + 1, int(invoice.number))
            last_number = invoice.number

    def test_sequence(self):
        c = InvoiceCreator(request=get_offline_request("testserver"))
        Invoice = c.get_invoice_model()
        order = Order.objects.first()
        number = c.generate_invoice_number(order)
        self.assertEqual(number, "000001")
        with transaction.atomic():
            c.create_invoice_model(
                order=order, legal_entity=c.get_legal_entity(), number=number
            )
        self.assertEqual(Invoice.objects.count(), 1)
        self.check_sequence(c)

        second_order = Order.objects.all()[1]
        with transaction.atomic():
            with self.assertRaises(
                IntegrityError,
                msg="Using the same order number should raise IntegrityError",
            ):
                c.create_invoice_model(
                    order=second_order, legal_entity=c.get_legal_entity(), number=number
                )
        self.assertEqual(Invoice.objects.count(), 1)
        self.check_sequence(c)

        second_order = Order.objects.all()[1]
        with transaction.atomic():
            with self.assertRaises(
                IntegrityError,
                msg="Using the same order number should raise IntegrityError",
            ):
                c.create_invoice_model(
                    order=second_order, legal_entity=c.get_legal_entity(), number=number
                )
        self.assertEqual(Invoice.objects.count(), 1)
        self.check_sequence(c)

        with transaction.atomic():
            number = c.generate_invoice_number(second_order)
            c.create_invoice_model(
                order=second_order, legal_entity=c.get_legal_entity(), number=number
            )
            self.assertEqual(number, "000002")
        self.assertEqual(Invoice.objects.count(), 2)
        self.check_sequence(c)

        third_order = Order.objects.all()[2]
        with transaction.atomic():
            number = c.generate_invoice_number(third_order)
            c.create_invoice_model(
                order=third_order, legal_entity=c.get_legal_entity(), number=number
            )
            self.assertEqual(number, "000003")
        self.assertEqual(Invoice.objects.count(), 3)
        self.check_sequence(c)

    def test_get_or_create_invoice_sequence(self):
        c = InvoiceCreator(request=get_offline_request("testserver"))
        order = Order.objects.all()[3]
        invoice = c.get_or_create_invoice(order)
        self.assertEqual(invoice.number, "000001")
        invoice.number = "000002"
        invoice.save()
        second_order = Order.objects.all()[1]
        with self.assertRaises(
            IntegrityError,
            msg="Using the same order number should raise IntegrityError",
        ):
            c.get_or_create_invoice(second_order)
        self.check_sequence(c)
        invoice.number = "000001"
        invoice.save()

        second_invoice = c.get_or_create_invoice(second_order)
        self.assertEqual(second_invoice.number, "000002")
        self.check_sequence(c)

    def test_invoice_get_legal_entity(self):
        invoice_creator = InvoiceCreator(request=get_offline_request("testserver"))
        legal_entity_no_order = invoice_creator.get_legal_entity()
        self.assertEqual(
            legal_entity_no_order.pk,
            1,
            "If no order is passed, it will take the first legal entity by default.",
        )

        order_1 = Order.objects.first()
        self.assertEqual(order_1.site.name, "example.com")
        legal_entity_order1 = invoice_creator.get_legal_entity(order=order_1)
        self.assertEqual(
            legal_entity_order1.shop_name,
            "example.com",
            "A order has been passed, it should find the legal entity with this site name.",
        )

        order_2 = Order.objects.get(pk=4)
        self.assertEqual(order_2.site.name, "shop2")
        legal_entity_order2 = invoice_creator.get_legal_entity(order=order_2)
        self.assertEqual(
            legal_entity_order2.shop_name,
            "shop2",
            "A order has been passed, it should find the legal entity with this site name.",
        )



# pylint: disable=no-name-in-module
from oscar.apps.communication import apps as oscar


class CommunicationConfig(oscar.CommunicationConfig):
    name = "ocyan.plugin.oscar_communications.communication"



# Generated by Django 2.2.13 on 2020-06-16 11:27

from django.db import migrations, transaction
import sequences


def forward(apps, schema_editor):
    Invoice = apps.get_model("oscar_invoices", "Invoice")
    invoices = Invoice.objects.select_for_update().order_by("pk")
    with transaction.atomic():
        last_invoice = invoices.last()
        if last_invoice is not None:
            sequences.get_next_value("oscar_invoices", initial_value=last_invoice.pk)


class Migration(migrations.Migration):

    dependencies = [
        ("oscar_invoices", "0001_initial"),
        ("sequences", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(forward, migrations.RunPython.noop),
    ]



from oscar.apps.communication import managers as oscar
from oscar.core.loading import get_class

EmailContext = get_class("order.context", "EmailContext")


class CommunicationTypeManager(oscar.CommunicationTypeManager):
    def get_and_render(self, code, context):
        if isinstance(context, EmailContext):
            return super().get_and_render(code, context)

        return super().get_and_render(code, EmailContext(**context))



from urllib.parse import urlunparse

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.contrib.sites.shortcuts import get_current_site
from django.template import Context, Template
from django.urls import NoReverseMatch, reverse

from ocyan.core.fender import config
from ocyan.main.loading import get_ocyan_classes

# pylint: disable=unbalanced-tuple-unpacking
(get_wagtailsettings, get_offline_request) = get_ocyan_classes(
    "wagtail_utils.utils", ["get_wagtailsettings", "get_offline_request"]
)


# pylint: disable=abstract-method
class AnonymousFullNameUser(AnonymousUser):
    def __init__(self, first_name, last_name, email):
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

    def get_full_name(self):
        """
        Return the first_name plus the last_name, with a space in between.
        """
        full_name = "%s %s" % (self.first_name or "", self.last_name or "")
        return full_name.strip()

    def get_short_name(self):
        """Return the short name for the user."""
        return self.first_name


class EmailContext(dict):
    def __init__(self, site=None, request=None, **kwargs):
        if site is not None and request is None:
            request = get_offline_request(site.domain)
        elif site is None and request is not None:
            site = get_current_site(request)

        super().__init__(site=site, request=request, **kwargs)

        self.populate_props()

    def __getattr__(self, attname):
        return self.get(attname)

    def populate_props(self):
        self.setdefault("settings", self.settings)

    @property
    def site(self):
        return self["site"]

    @property
    def settings(self):
        return get_wagtailsettings(self.site.domain)


class OrderContext(EmailContext):
    def __init__(self, order, **kwargs):
        kwargs.setdefault("request", get_offline_request(order.site.domain))
        kwargs.setdefault("site", order.site)
        super().__init__(order=order, **kwargs)

    def populate_props(self):
        super().populate_props()
        self.setdefault("user", self.user)
        self.setdefault("lines", self.lines)
        self.setdefault("status_url", self.status_url)
        self.setdefault(
            "static_base_url", getattr(settings, "OSCAR_STATIC_BASE_URL", None)
        )
        self.setdefault("user", self.user)

    @property
    def user(self):
        _user = self.order.user
        _shipping_address = self.order.shipping_address

        # try to fill in user details for anonymous orders
        if self.order.is_anonymous and _shipping_address:
            return AnonymousFullNameUser(
                _shipping_address.first_name,
                _shipping_address.last_name,
                self.order.guest_email,
            )
        elif (
            _user
        ):  # this is not an anonymous order, but try to complete unfilled user details
            if _user.last_name or _user.first_name:
                return _user
            elif _shipping_address:
                _user.last_name = _shipping_address.last_name
                _user.first_name = _shipping_address.first_name
                return _user

            return _user
        elif _shipping_address:  # user has deleted profile.
            return AnonymousFullNameUser(
                _shipping_address.first_name,
                _shipping_address.last_name,
                self.order.guest_email,
            )

        # we know absolutely nothing about this person
        return AnonymousUser()

    @property
    def lines(self):
        return self.order.lines.all()

    @property
    def status_url(self):
        if not self.user.is_authenticated:
            # Attempt to add the anon order status URL to the email template
            # ctx.
            try:
                path = reverse(
                    "customer:anon-order",
                    kwargs={
                        "order_number": self.order.number,
                        "hash": self.order.verification_hash(),
                    },
                )
                return urlunparse(("https", self.site.domain, path, None, None, None))
            except NoReverseMatch:
                # We don't care that much if we can't resolve the URL
                return None

        return None


class ShippingContext(OrderContext):
    def __init__(self, shipping_event, **kwargs):
        self.shipping_event = shipping_event
        super().__init__(shipping_event.order, **kwargs)

    def populate_props(self):
        super().populate_props()
        self.setdefault("tracking_url", self.tracking_url)

        # hasattr because tracking is OneToOneField
        if hasattr(self.shipping_event, "tracking"):
            self.setdefault("tracking", self.shipping_event.tracking)

    @property
    def tracking_url(self):
        # hasattr because tracking is OneToOneField
        if hasattr(self.shipping_event, "tracking"):
            return self.shipping_event.tracking.get_tracking_url
        elif self.shipping_event.notes:
            maybe_an_url = self.shipping_event.notes
            tracking_url_template = config.get(
                "oscar_communications", "tracking_url_template", None
            )

            if maybe_an_url.startswith("http"):
                return maybe_an_url
            elif tracking_url_template is not None:
                tracking_template = Template(tracking_url_template)
                return tracking_template.render(Context({"code": maybe_an_url}))

        return None

    @property
    def lines(self):
        return [
            (lq.line, lq.quantity) for lq in self.shipping_event.line_quantities.all()
        ]



# pylint: disable=unused-import
import logging

from django.db import transaction
from django.utils.translation import gettext as _

from oscar.apps.order import exceptions as order_exceptions
from oscar.core.loading import get_class, get_classes, get_model

from ocyan.core.fender import config
from ocyan.main.loading import get_baseclass
from ocyan.plugin.oscar.info import OSCAR_NEEDS_OXYAN
from ocyan.plugin.oscar.pipeline import (
    PIPELINE_STATUS_CANCELLED,
    PIPELINE_STATUS_COMPLETE,
    PIPELINE_STATUS_DELAYED_PAYMENT,
    PIPELINE_STATUS_NEW,
    PIPELINE_STATUS_PAID,
    PIPELINE_STATUS_PENDING_PAYMENT,
    PIPELINE_STATUS_REFUND,
    PIPELINE_STATUS_RESTOCKED,
)
from ocyan.plugin.oscar_shipping.plugins import get_shipping_event_types
from ocyan.plugin.oscar_shipping.shipping.constants import SHIPPING_EVENT_SHIPPED

BaseEventHandler = get_baseclass(
    "order.processing",
    "EventHandler",
    "ocyan.plugin.oscar_communications.order_extension",
)
send_message = get_class("order.shortcuts", "send_message")
send_order_confirmation_message, send_new_order_status_message = get_classes(
    "order.communications",
    ["send_order_confirmation_message", "send_new_order_status_message"],
)
ShippingContext = get_class("order.context", "ShippingContext")
ShippingEventQuantity = get_model("order", "ShippingEventQuantity")
Tracking = get_model("order", "Tracking")

if OSCAR_NEEDS_OXYAN:
    CommunicationEventType = get_model("communication", "CommunicationEventType")
else:
    CommunicationEventType = get_model("customer", "CommunicationEventType")

logger = logging.getLogger(__name__)


class EventHandler(BaseEventHandler):
    def handle_order_status_change(self, order, new_status, note_msg=None):
        with transaction.atomic():
            super(EventHandler, self).handle_order_status_change(
                order, new_status, note_msg
            )
            if new_status in [PIPELINE_STATUS_PAID, PIPELINE_STATUS_DELAYED_PAYMENT]:
                send_order_confirmation_message("ORDER_PLACED", order)
            elif new_status == PIPELINE_STATUS_CANCELLED:
                self.cancel_stock_allocations(order)

            # Search for a communication event specific for the new order status
            status_event_code = "ORDER_STATUS_%s" % new_status.upper()
            status_event_exists = CommunicationEventType.objects.filter(
                code=status_event_code
            ).exists()
            if status_event_exists:
                send_new_order_status_message(status_event_code, order)

    # pylint: disable=unused-argument
    def can_consume_stock_allocations(self, order, event_type, lines, line_quantities):
        shipping_event_types = get_shipping_event_types()

        if event_type.code not in shipping_event_types:
            # we only consume allocations for shipping events that are registered with
            # contribute_to_shipping_event_types
            return False

        shipping_events = order.shipping_events.filter(
            event_type__code__in=shipping_event_types
        )
        if not shipping_events.exists():
            # there has been no allocation at all yet, because no shipping events
            # that trigger allocations have been registered yet
            return True
        else:
            # there allready are shipping events.
            # we use order._is_event_complete
            # to check if the quantities for all shipping event types of type
            # SHIPPING_EVENT_TYPES sum to the total order quantities.
            return not order._is_event_complete(  # pylint: disable=protected-access
                ShippingEventQuantity.objects.filter(
                    event__event_type__code__in=shipping_event_types
                )
            )

    def handle_shipping_event(
        self, order, event_type, lines, line_quantities, **kwargs
    ):
        # consuming stock and shipping event has to be tied together to in a
        # transaction to avoid double allocations in case of errors.
        with transaction.atomic():
            do_consume_stock_allocations = self.can_consume_stock_allocations(
                order, event_type, lines, line_quantities
            )

            # check values and create shipping event
            event = super(EventHandler, self).handle_shipping_event(
                order, event_type, lines, line_quantities, **kwargs
            )

            if do_consume_stock_allocations:
                # seems everything is valid, so consume stock because we took the
                # product and stuck it in a package.
                self.consume_stock_allocations(order, lines, line_quantities)

        if event_type.code in get_shipping_event_types():
            # put all fully shipped lines to status completed
            for line in lines:
                if self.have_lines_passed_shipping_event(
                    order, [line], [line.quantity], event_type
                ):
                    try:
                        line.set_status(PIPELINE_STATUS_COMPLETE)
                    except order_exceptions.InvalidLineStatus:
                        pass  # shut this up, ne need to display error

            # send shipping email if needed.
            if config.get("oscar_communications", "send_shipping_email", False):
                ctx = ShippingContext(event)
                send_message(ctx, SHIPPING_EVENT_SHIPPED)

            # if there are nomore lines that are uncompleted, complete order.
            if not order.lines.exclude(status=PIPELINE_STATUS_COMPLETE).exists():
                try:
                    msg = _("All lines have shipped")
                    self.handle_order_status_change(
                        order, PIPELINE_STATUS_COMPLETE, msg
                    )
                except order_exceptions.InvalidOrderStatus:
                    pass  # shut this up, ne need to display error

        return event

    def create_payment_event(
        self, order, event_type, amount, lines=None, line_quantities=None, **kwargs
    ):
        payment_event = super().create_payment_event(
            order, event_type, amount, lines, line_quantities, **kwargs
        )
        if payment_event.event_type.code == PIPELINE_STATUS_REFUND:
            # update the balance and refund the specified amount
            source = order.sources.first()
            if source is not None:
                source.refund(payment_event.amount)

            # If lines have been fully refunded update the line status to refund
            for line in payment_event.lines.all():
                if line.payment_event_quantity(event_type) == line.quantity:
                    line.set_status(PIPELINE_STATUS_REFUND)

    def cancel_stock_allocations(self, order, lines=None, line_quantities=None):
        if not lines:
            lines = order.lines.exclude(status=PIPELINE_STATUS_RESTOCKED)
        if not line_quantities:
            line_quantities = [line.quantity for line in lines]

        # everything has been allocated zo don't do anything
        if not line_quantities:
            return

        super().cancel_stock_allocations(order, lines, line_quantities)
        for line in lines:
            line.set_status(PIPELINE_STATUS_RESTOCKED)

    def validate_shipping_event(
        self, order, event_type, lines, line_quantities, **kwargs
    ):
        if self.are_stock_allocations_available(lines, line_quantities):
            return super(EventHandler, self).validate_shipping_event(
                order, event_type, lines, line_quantities, **kwargs
            )
        else:
            msg = _(
                "Not enough stock or allocations available, please update stockrecords first."
            )
            raise order_exceptions.InvalidShippingEvent(msg)

    def create_shipping_event(
        self, order, event_type, lines, line_quantities, **kwargs
    ):
        shipping_event = super().create_shipping_event(
            order, event_type, lines, line_quantities, **kwargs
        )
        self.create_tracking(shipping_event, **kwargs)
        return shipping_event

    def create_tracking(
        self,
        shipping_event,
        carrier=None,
        tracking_code=None,
        tracking_url=None,
        **kwargs,
    ):
        """
        Requires carrier and tracking_code or tracking_url in kwargs for it to work.
        """
        if carrier:
            return Tracking.create_for_shipping_event(
                shipping_event, carrier, tracking_code, tracking_url, **kwargs
            )



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse

from oscar.core.loading import get_model

from ocyan.plugin.oscar.pipeline import (
    PIPELINE_STATUS_CANCELLED,
    PIPELINE_STATUS_PAID,
    PIPELINE_STATUS_PENDING_PAYMENT,
    PIPELINE_STATUS_REFUND,
    PIPELINE_STATUS_RESTOCKED,
)

User = get_user_model()
Order = get_model("order", "Order")


def change_line_status(status):
    return {
        "selected_line": "1",
        "selected_line_qty_1": "1",
        "selected_line_qty_2": "1",
        "line_action": "change_line_statuses",
        "new_status": status,
        "shipping_event_type": "",
        "reference": "",
        "payment_event_type": "",
        "amount": "",
    }


def change_order_status(status):
    return {"new_status": status, "order_action": "change_order_status"}


def refund(amount):
    return {
        "selected_line": "1",
        "selected_line_qty_1": "1",
        "selected_line_qty_2": "1",
        "new_status": "",
        "shipping_event_type": "",
        "reference": "",
        "line_action": "create_payment_event",
        "payment_event_type": PIPELINE_STATUS_REFUND,
        "amount": amount,
    }


class TestOrderStatus(TestCase):
    fixtures = [
        "communications/auth",
        "communications/address",
        "communications/catalogue",
        "communications/order",
        "communications/basket",
        "communications/customer",
        "communications/invoices",
    ]

    def setUp(self):
        super().setUp()
        user = User.objects.get(is_superuser=True)
        self.client.force_login(user)

    def order_under_test(self):
        return Order.objects.get(pk=1)

    def assert_initial_status(self, order):
        order.refresh_from_db()
        self.assertEqual(
            order.lines.count(), 2, "there should be 2 lines on the order under test"
        )

        stockrecord1, stockrecord2 = order.lines.values_list("stockrecord", flat=True)
        self.assertNotEqual(
            stockrecord1, stockrecord2, "The lines should have different stockrecords"
        )

        status1, status2 = order.lines.values_list("status", flat=True)
        self.assertEqual(status1, status2, "The lines should have the same status")

        line1, line2 = order.lines.all()
        stockrecord1, stockrecord2 = line1.stockrecord, line2.stockrecord

        self.assertEqual(stockrecord1.num_in_stock, 34)
        self.assertEqual(stockrecord1.num_allocated, 3)
        self.assertEqual(stockrecord1.net_stock_level, 31)

        self.assertEqual(stockrecord2.num_in_stock, 345)
        self.assertEqual(stockrecord2.num_allocated, 2)
        self.assertEqual(stockrecord2.net_stock_level, 343)

        source = order.sources.first()
        self.assertEqual(source.balance, 0)

    def assert_deallocation(self, order):
        order.refresh_from_db()
        line1, line2 = order.lines.all()
        stockrecord1, stockrecord2 = line1.stockrecord, line2.stockrecord

        self.assertEqual(stockrecord1.num_in_stock, 34)
        self.assertEqual(
            stockrecord1.num_allocated,
            2,
            "After cancelling the order, there should be one less allocated item",
        )
        self.assertEqual(
            stockrecord1.net_stock_level,
            32,
            "After cancelling the order, there should be one more product available",
        )

        self.assertEqual(stockrecord2.num_in_stock, 345)
        self.assertEqual(
            stockrecord2.num_allocated,
            1,
            "After cancelling the order, there should be one less allocated item",
        )
        self.assertEqual(
            stockrecord2.net_stock_level,
            344,
            "After cancelling the order, there should be one more product available",
        )

    def test_parameters_are_correct(self):
        order = self.order_under_test()
        status1, _ = order.lines.values_list("status", flat=True)
        self.assertEqual(
            status1, PIPELINE_STATUS_PAID, "The lines should have status paid"
        )

        self.assert_initial_status(order)

    def test_refund_status_only(self):
        order = self.order_under_test()

        url = reverse("dashboard:order-detail", kwargs={"number": order.number})
        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_REFUND), follow=True
        )
        self.assertEqual(response.status_code, 200)

        self.assert_initial_status(order)

    def test_cancel_whole_order(self):
        "When an orderis being cancelled, all stock should be unallocated."
        order = self.order_under_test()

        url = reverse("dashboard:order-detail", kwargs={"number": order.number})
        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_CANCELLED), follow=True
        )
        self.assertEqual(response.status_code, 200)

        # you can't cancel order in one go, they should be refunded before cancelling
        for message in response.context["messages"]:
            self.assertEqual(message.message, "Ongeldige formulier invoer")
            self.assertEqual(message.level, 40)

        # nothing should have happened
        self.assert_initial_status(order)

        # now refund first before cancelling
        self.test_refund_status_only()
        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_CANCELLED), follow=True
        )
        self.assertEqual(response.status_code, 200)

        self.assert_deallocation(order)

    def test_cancel_pending_cancel_whole_order(self):
        "When an orderis being cancelled, all stock should be unallocated."
        order = self.order_under_test()

        url = reverse("dashboard:order-detail", kwargs={"number": order.number})
        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_CANCELLED), follow=True
        )
        self.assertEqual(response.status_code, 200)

        # you can't cancel order in one go, they should be refunded before cancelling
        for message in response.context["messages"]:
            self.assertEqual(message.message, "Ongeldige formulier invoer")
            self.assertEqual(message.level, 40)

        # nothing should have happened
        self.assert_initial_status(order)

        line1, line2 = order.lines.all()
        stockrecord1, stockrecord2 = line1.stockrecord, line2.stockrecord

        self.assertEqual(stockrecord1.num_allocated, 3)
        self.assertEqual(stockrecord2.num_allocated, 2)

        # now refund first before cancelling
        self.test_refund_status_only()
        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_CANCELLED), follow=True
        )
        self.assertEqual(response.status_code, 200)

        stockrecord1.refresh_from_db()
        stockrecord2.refresh_from_db()

        self.assertEqual(stockrecord1.num_allocated, 2)
        self.assertEqual(stockrecord2.num_allocated, 1)

        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_PENDING_PAYMENT), follow=True
        )
        self.assertEqual(response.status_code, 200)

        stockrecord1.refresh_from_db()
        stockrecord2.refresh_from_db()
        self.assertEqual(stockrecord1.num_allocated, 2)
        self.assertEqual(stockrecord2.num_allocated, 1)

        response = self.client.post(
            url, change_order_status(PIPELINE_STATUS_CANCELLED), follow=True
        )
        self.assertEqual(response.status_code, 200)

        stockrecord1.refresh_from_db()
        stockrecord2.refresh_from_db()
        self.assertEqual(stockrecord1.num_allocated, 2)
        self.assertEqual(stockrecord2.num_allocated, 1)

        self.assert_deallocation(order)

    def test_refund_order_line(self):
        order = self.order_under_test()

        self.assert_initial_status(order)

        url = reverse("dashboard:order-detail", kwargs={"number": order.number})
        response = self.client.post(url, refund(12), follow=True)
        self.assertEqual(response.status_code, 200)

        source = order.sources.first()
        self.assertEqual(
            source.balance,
            12,
            "There should now be open balance on the order due to the refund",
        )

        line1, line2 = order.lines.all()
        self.assertEqual(line1.status, PIPELINE_STATUS_REFUND)
        self.assertEqual(line2.status, PIPELINE_STATUS_PAID)

    def test_cancel_order_line(self):
        order = self.order_under_test()

        url = reverse("dashboard:order-detail", kwargs={"number": order.number})

        self.test_refund_order_line()
        response = self.client.post(
            url, change_line_status(PIPELINE_STATUS_CANCELLED), follow=True
        )
        self.assertEqual(response.status_code, 200)

        line1, line2 = order.lines.all()
        self.assertEqual(line1.status, PIPELINE_STATUS_RESTOCKED)
        self.assertEqual(line2.status, PIPELINE_STATUS_PAID)

        stockrecord1, stockrecord2 = line1.stockrecord, line2.stockrecord

        self.assertEqual(stockrecord2.num_in_stock, 345)
        self.assertEqual(stockrecord2.num_allocated, 2)
        self.assertEqual(stockrecord2.net_stock_level, 343)

        self.assertEqual(stockrecord1.num_in_stock, 34)
        self.assertEqual(
            stockrecord1.num_allocated,
            2,
            "After cancelling the order, there should be one less allocated item",
        )
        self.assertEqual(
            stockrecord1.net_stock_level,
            32,
            "After cancelling the order, there should be one more product available",
        )



# -*- coding: utf-8 -*-
default_app_config = (
    "ocyan.plugin.oscar_communications.apps.OcyanPluginOscarCommunicationsAppConfig"
)
from ocyan.plugin.oscar.info import OSCAR_NEEDS_OXYAN

if OSCAR_NEEDS_OXYAN:
    from .entrypoint import Oscar3CommunicationsConfig  # noqa
else:
    from .entrypoint import OscarCommunicationsConfig  # noqa



from django.conf import settings
from django.db import connection
from django.utils.translation import override
from uwsgidecoratorsfallback import spool

from oscar.core.loading import get_class, get_model

from ocyan.core.fender import config
from ocyan.main.loading import get_ocyan_class

get_offline_request = get_ocyan_class("wagtail_utils.utils", "get_offline_request")
send_message = get_class("order.shortcuts", "send_message")
OrderContext = get_class("order.context", "OrderContext")

try:
    InvoiceCreator = get_class("oscar_invoices.utils", "InvoiceCreator")
except LookupError:
    pass

Order = get_model("order", "Order")
PaymentMethod = get_model("checkout", "PaymentMethod")


@spool
def send_order_confirmation_message_async(args):
    # the spooler is forked so the db connection might be closed
    try:
        if not connection.is_usable():
            connection.close()
    except AttributeError:
        connection.close_if_unusable_or_obsolete()

    event_type = args["event_type"]
    order_id = args["order_id"]
    order = Order.objects.get(pk=order_id)
    request = get_offline_request(order.site.domain)

    with override(order.language_code):
        kwargs = {}
        if config.get("oscar_communications", "invoice", False):
            send_invoice = True
            if order.payment_source_type is not None:
                payment_method = PaymentMethod.objects.filter(
                    key=order.payment_source_type.code
                ).first()
                if payment_method is not None:
                    send_invoice = payment_method.send_invoice

            if send_invoice:
                ic = InvoiceCreator(request=request)
                invoice = ic.get_or_create_invoice(order)
                if invoice is not None:
                    kwargs["attachments"] = [invoice.document]

        ctx = OrderContext(order)
        send_message(ctx, event_type, bcc_admins=True, **kwargs)


@spool
def send_new_order_status_message_async(args):
    try:
        if not connection.is_usable():
            connection.close()
    except AttributeError:
        connection.close_if_unusable_or_obsolete()

    event_type = args["event_type"]
    order_id = args["order_id"]
    order = Order.objects.get(pk=order_id)
    order_status = "ORDER_STATUS_%s" % order.status.upper()
    bcc_admins = getattr(settings, "%s_BCC_ADMINS" % order_status, False)
    attach_invoice = getattr(settings, "%s_ATTACH_INVOICE" % order_status, False)

    with override(order.language_code):
        kwargs = {}

        if attach_invoice:
            request = get_offline_request(order.site.domain)
            ic = InvoiceCreator(request=request)
            invoice = ic.get_or_create_invoice(order)
            if invoice is not None:
                kwargs["attachments"] = [invoice.document]

        ctx = OrderContext(order)
        send_message(ctx, event_type, bcc_admins=bcc_admins, **kwargs)



# there is no need to use get_class for the spooler function because you
# can not override it. There cannot be 2 spooler functions in uwsgi with the
# same name, which is what you would end up with if you'd override the spooler
# function in an app extension.
import ocyan.plugin.oscar.signals

from .spooler import (
    send_new_order_status_message_async,
    send_order_confirmation_message_async,
)


def send_order_confirmation_message(event_type, order):
    send_order_confirmation_message_async.spool(
        event_type=event_type, order_id=str(order.pk)
    )
    ocyan.plugin.oscar.signals.order_ready.send(sender=order.__class__, order=order)


def send_new_order_status_message(event_type, order):
    send_new_order_status_message_async.spool(
        event_type=event_type, order_id=str(order.pk)
    )



from django.utils.translation import gettext_lazy as _

from oscar.apps.customer.alerts.views import (
    ProductAlertCreateView as BaseProductAlertCreateView,
)


class ProductAlertCreateView(BaseProductAlertCreateView):
    def form_valid(self, form):
        error_field = "__all__" if form.user else "email"
        form.add_error(
            error_field,
            _(
                "Because of spammers abusing this form, unfortunately at this point we cannot accept product alert requests"
            ),
        )
        return self.form_invalid(form)



# -*- coding: utf-8 -*-
import logging

from django.apps import apps
from django.conf.urls.i18n import i18n_patterns
from django.urls import include, path

from ocyan.core.fender import config
from ocyan.core.plugins import OcyanPluginConfig
from ocyan.core.utils import merge_dicts
from ocyan.plugin.oscar.pipeline import PIPELINE_STATUS_COMPLETE
from ocyan.plugin.oscar.settings import OSCAR_ORDER_STATUS_CASCADE

logger = logging.getLogger(__name__)


class OscarCommunicationsConfig(OcyanPluginConfig):
    def contribute_to_installed_apps(self, installed_apps):
        installed_apps.django.append("sequences.apps.SequencesConfig")
        installed_apps.ocyan.append("oscar_invoices")
        installed_apps.ocyan.append("ocyan.plugin.oscar_communications")
        installed_apps.ocyan.append(
            "ocyan.plugin.oscar_communications.oscar_invoices_extension",
            extends="oscar_invoices",
        )

        installed_apps.oscar.append(
            "ocyan.plugin.oscar_communications.customer_extension",
            extends="oscar.apps.customer",
        )
        installed_apps.oscar.register_extension(
            "ocyan.plugin.oscar_communications.customer_extension",
            "ocyan.plugin.oscar_multi_site.customer_alerts_ms_extension",
        )
        installed_apps.oscar.register_extension(
            "ocyan.plugin.oscar_communications.order_extension",
            "ocyan.plugin.oscar_order.order",
        )
        return installed_apps

    def contribute_to_settings(self, settings):
        # we will remove the cascading of PIPELINE_STATUS_COMPLETE from the
        # order to the lines, because it will be handled by shipping events in
        # this plugin
        try:
            # modify the setting by it's reference directly
            del OSCAR_ORDER_STATUS_CASCADE[PIPELINE_STATUS_COMPLETE]
        except KeyError:
            pass

        return settings

    def _get_invoice_url_target(self):
        invoice_app = apps.get_app_config("oscar_invoices")
        invoice_urls = getattr(invoice_app, "urls", None)
        if invoice_urls is not None:
            return invoice_urls

        get_urls = getattr(invoice_app, "get_urls", None)
        if callable(get_urls):
            namespace = getattr(invoice_app, "namespace", invoice_app.label)
            return include((get_urls(), invoice_app.label), namespace=namespace)

        raise AttributeError(
            "oscar_invoices app config does not expose urls/get_urls"
        )

    def contribute_to_urlpatterns(self, urlpatterns):
        new_patterns = []
        try:
            new_patterns.append(
                path(r"dashboard/oscar-invoices/", self._get_invoice_url_target())
            )
        except (LookupError, AttributeError) as exc:
            logger.warning("Skipping oscar_invoices dashboard urls: %s", exc)
        if config.i18n_enabled:
            new_patterns += i18n_patterns(
                path(
                    r"admin/communications/",
                    include("ocyan.plugin.oscar_communications.urls"),
                ),
                prefix_default_language=False,
            )
        else:
            new_patterns.append(
                path(
                    r"admin/communications/",
                    include("ocyan.plugin.oscar_communications.urls"),
                )
            )

        new_patterns.extend(list(urlpatterns))
        return new_patterns

    def contribute_to_fixtures(self, fixtures):
        fixtures.add("invoices", "communications_demo_data/invoices")
        return fixtures

    def get_spooler_modules(self, modules):
        modules.append("ocyan.plugin.oscar_communications.order_extension.spooler")
        return modules


class Oscar3CommunicationsConfig(OscarCommunicationsConfig):
    def contribute_to_installed_apps(self, installed_apps):
        installed_apps = super().contribute_to_installed_apps(installed_apps)
        installed_apps.oscar.append(
            "ocyan.plugin.oscar_communications.communication.apps.CommunicationConfig",
            replaces="oscar.apps.communication.apps.CommunicationConfig",
        )
        installed_apps.oscar.register_extension(
            "ocyan.plugin.oscar_communications.customer_extension",
            "oscar.apps.communication",
        )
        installed_apps.oscar.register_extension(
            "ocyan.plugin.oscar_communications.oxyan_communications_extension",
            "ocyan.plugin.oscar_communications.customer_extension",
        )
        return installed_apps

    def contribute_to_settings(self, settings):
        return merge_dicts(
            settings,
            {
                "MIGRATION_MODULES": {
                    "communication": "ocyan.plugin.oscar_communications.communication.migrations"
                },
            },
        )



from django.dispatch import receiver

from oscar.apps.order.signals import order_line_status_changed
from oscar.core.loading import get_class

from ocyan.plugin.oscar.pipeline import PIPELINE_STATUS_CANCELLED

EventHandler = get_class("order.processing", "EventHandler")


# pylint: disable=unused-argument
@receiver(order_line_status_changed)
def handle_order_line_status_changed(line, old_status, new_status, **kwargs):
    if old_status != new_status and new_status == PIPELINE_STATUS_CANCELLED:
        EventHandler().cancel_stock_allocations(line.order, [line], [line.quantity])



# -*- coding: utf-8 -*-
import unittest
from io import BytesIO
from unittest import mock

import django.core.mail as email
from django.core.files import File
from django.test import TestCase, override_settings

from oscar.core.loading import get_model

import ocyan.core.fender
from ocyan.plugin.oscar.pipeline import PIPELINE_STATUS_PAID

from ocyan.plugin.oscar_communications.order_extension.processing import EventHandler

Order = get_model("order", "Order")

BCC_EMAIL = {
    "django": {"email_to": "henk@nub.lol"},
    "oscar_communications": {
        "invoice": True,
        "send_shipping_email": True,
        "tracking_url_template": "http://posnl.nl/{{ code }}?activate_my_dick",
    },
}

invoice_creator_instance = mock.MagicMock()
attrs = {
    "name": "klaas.pdf",
    "document": File(BytesIO(b""), name="klaas.pdf"),
    "get_or_create_invoice.return_value": invoice_creator_instance,
}
invoice_creator_instance.configure_mock(**attrs)
InvoiceCreator = mock.MagicMock(return_value=invoice_creator_instance)


@override_settings(DEFAULT_TO_EMAIL=["henk@nub.lol"])
@mock.patch.object(ocyan.core.fender.config, "settings", BCC_EMAIL)
@mock.patch(
    "ocyan.plugin.oscar_communications.order_extension.spooler.InvoiceCreator",
    InvoiceCreator,
    create=True,
)
class CommunicationsTest(TestCase):
    fixtures = [
        "communications/auth",
        "communications/address",
        "communications/catalogue",
        "communications/order",
        "communications/basket",
        "communications/customer",
    ]

    def setUp(self):
        self.evt_handler = EventHandler()
        setattr(email, "outbox", [])

    def test_invoice_registered_user(self):
        with self.settings(
            EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend"
        ):
            registered_order = Order.objects.get(pk=1)
            self.assertEqual(len(email.outbox), 0)
            self.evt_handler.handle_order_status_change(
                registered_order, PIPELINE_STATUS_PAID, note_msg="Test test"
            )
            self.assertEqual(len(email.outbox), 1)
            mail = email.outbox[0]
            self.assertEqual(mail.to, ["admin@admin.admin"])
            self.assertEqual(mail.attachments, [("klaas.pdf", b"", "application/pdf")])
            self.assertEqual(
                mail.recipients(),
                ["admin@admin.admin", "henk@nub.lol"],
                "The order mail should be sent to admin as well as customer",
            )
            self.assertEqual(
                mail.subject,
                "Klaas Berten admin@admin.admin example.com 2",
                "The subject should contain all the promoted variables",
            )
            self.assertTrue(
                mail.body.strip().startswith("Beste Klaas Berten,"),
                "The user should be greeted with his user.get_full_name",
            )
            self.assertTrue(
                mail.body.strip().endswith("Het example.com team"),
                "The message should conclude properly",
            )

            self.assertIn("Hallo Klaas Berten", str(mail.alternatives))
            self.assertIn("Zoekeenhuisstraat 3", str(mail.alternatives))
            self.assertIn("Het example.com team", str(mail.alternatives))

    def test_invoice_registered_user_with_billing_address(self):
        with self.settings(
            EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend"
        ):
            registered_order = Order.objects.get(pk=3)
            self.assertEqual(len(email.outbox), 0)
            self.evt_handler.handle_order_status_change(
                registered_order, PIPELINE_STATUS_PAID, note_msg="Test test"
            )
            self.assertEqual(len(email.outbox), 2)
            billing_mail = email.outbox[0]
            mail = email.outbox[1]
            self.assertEqual(billing_mail.to, ["viggo@mandelblog.com"])
            self.assertEqual(mail.to, ["admin@admin.admin"])
            self.assertEqual(
                billing_mail.attachments, [("klaas.pdf", b"", "application/pdf")]
            )
            self.assertEqual(mail.attachments, [])
            self.assertEqual(
                mail.recipients(),
                ["admin@admin.admin", "henk@nub.lol"],
                "The order mail should be sent to admin as well as customer",
            )
            self.assertEqual(
                mail.subject,
                "Klaas Berten admin@admin.admin example.com 6",
                "The subject should contain all the promoted variables",
            )
            self.assertTrue(
                mail.body.strip().startswith("Beste Klaas Berten,"),
                "The user should be greeted with his user.get_full_name",
            )
            self.assertTrue(
                mail.body.strip().endswith("Het example.com team"),
                "The message should conclude properly",
            )

            self.assertIn("Hallo Klaas Berten", str(mail.alternatives))
            self.assertIn("Zoekeenhuisstraat 3", str(mail.alternatives))
            self.assertIn("Het example.com team", str(mail.alternatives))

    def test_invoice_registered_user_with_billing_address_same_email(self):
        with self.settings(
            EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend"
        ):
            registered_order = Order.objects.get(pk=4)
            self.assertEqual(len(email.outbox), 0)
            self.evt_handler.handle_order_status_change(
                registered_order, PIPELINE_STATUS_PAID, note_msg="Test test"
            )
            self.assertEqual(len(email.outbox), 1)
            mail = email.outbox[0]
            self.assertEqual(mail.to, ["admin@admin.admin"])
            self.assertEqual(mail.attachments, [("klaas.pdf", b"", "application/pdf")])
            self.assertEqual(
                mail.recipients(),
                ["admin@admin.admin", "henk@nub.lol"],
                "The order mail should be sent to admin as well as customer",
            )
            self.assertEqual(
                mail.subject,
                "Klaas Berten admin@admin.admin shop2 8",
                "The subject should contain all the promoted variables",
            )
            self.assertTrue(
                mail.body.strip().startswith("Beste Klaas Berten,"),
                "The user should be greeted with his user.get_full_name",
            )
            self.assertTrue(
                mail.body.strip().endswith("Het shop2 team"),
                "The message should conclude properly",
            )

            self.assertIn("Hallo Klaas Berten", str(mail.alternatives))
            self.assertIn("Zoekeenhuisstraat 3", str(mail.alternatives))
            self.assertIn("Het shop2 team", str(mail.alternatives))

    def test_invoice_anonymous_user(self):
        with self.settings(
            EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend"
        ):
            registered_order = Order.objects.get(pk=2)
            self.assertEqual(len(email.outbox), 0)
            self.evt_handler.handle_order_status_change(
                registered_order, PIPELINE_STATUS_PAID, note_msg="Test test"
            )
            self.assertEqual(len(email.outbox), 1)
            mail = email.outbox[0]
            self.assertEqual(mail.to, ["piet@example.com"])
            self.assertEqual(mail.attachments, [("klaas.pdf", b"", "application/pdf")])
            self.assertEqual(
                mail.recipients(),
                ["piet@example.com", "henk@nub.lol"],
                "The order mail should be sent to admin as well as customer",
            )
            self.assertEqual(
                mail.subject,
                "Piet Henken piet@example.com example.com 3",
                "The subject should contain all the promoted variables",
            )
            self.assertTrue(
                mail.body.strip().startswith("Beste Piet Henken,"),
                "The user should be greeted with his user.get_full_name",
            )
            self.assertTrue(
                mail.body.strip().endswith("Het example.com team"),
                "The message should conclude properly",
            )

            self.assertIn("Hallo Piet Henken", str(mail.alternatives))
            self.assertIn("zultieboi", str(mail.alternatives))
            self.assertIn("Het example.com team", str(mail.alternatives))

    @unittest.skip("geen zin meer in")
    def test_shipping_registered_user(self):
        self.test_invoice_registered_user()

    @unittest.skip("geen zin meer in")
    def test_shipping_anonymous_user(self):
        self.test_invoice_anonymous_user()


class CommunicationEventTest(TestCase):
    fixtures = [
        "communications/auth",
        "communications/address",
        "communications/catalogue",
        "communications/order",
        "communications/basket",
        "communications/customer",
        "communications/allreadysent",
    ]

    def setUp(self):
        self.evt_handler = EventHandler()
        setattr(email, "outbox", [])

    def test_idempotent_communication_event(self):
        with self.settings(
            EMAIL_BACKEND="django.core.mail.backends.locmem.EmailBackend"
        ):
            registered_order = Order.objects.get(pk=1)
            self.assertEqual(len(email.outbox), 0)
            self.evt_handler.handle_order_status_change(
                registered_order, PIPELINE_STATUS_PAID, note_msg="Test test"
            )
            self.assertEqual(len(email.outbox), 0)



import logging
from copy import deepcopy
from unittest import mock

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import OperationalError, transaction
from django.template import base as template
from django.template import defaulttags
from django.utils.translation import override

from wagtail.contrib.settings.templatetags import wagtailsettings_tags
from wagtail.images.templatetags import wagtailimages_tags

from oscar.core.loading import get_class, get_model

from ocyan.core.fender import config
from ocyan.main.templatetags import ocyan_main
from ocyan.plugin.oscar.info import OSCAR_NEEDS_OXYAN

logger = logging.getLogger(__name__)

if OSCAR_NEEDS_OXYAN:
    CommunicationEventType = get_model("communication", "CommunicationEventType")
else:
    CommunicationEventType = get_model("customer", "CommunicationEventType")

EmailContext = get_class("order.context", "EmailContext")


# override forloop registration
def do_not_change_anything(parser, token):  # pylint: disable=unused-argument
    return template.TextNode("{%s}" % "% {} %".format(token.contents))


def execute_and_rerender(func):
    def _wrapped(parser, token):
        func(parser, token)
        return do_not_change_anything(parser, token)

    return _wrapped


# override filters, leave them unchanged
class NoFilterExpression(template.FilterExpression):
    def resolve(self, context, ignore_failures=False):
        if "|" in self.token:
            return "{{ %s }}" % self.token
        return super().resolve(context)


# override parse so filters will not be changed
class NoFilterParser(template.Parser):
    def compile_filter(self, token):
        return NoFilterExpression(token, self)


# make sure variables are rendered unchanged.
class DotPath(object):
    def __init__(self, *names):
        self.names = names

    def __getitem__(self, name):
        return DotPath(*self.names, name)

    def __str__(self):
        return "{{ %s }}" % ".".join(self.names)


class Command(BaseCommand):
    help = "Populate the email communications in the dashboard"

    def add_arguments(self, parser):
        parser.add_argument(
            "event_types", nargs="+", help="CommunicationEventType code names"
        )
        parser.add_argument(
            "--language-code",
            "-l",
            default=settings.LANGUAGE_CODE,
            help="language code to activate",
        )
        parser.add_argument(
            "--variable",
            action="append",
            type=str,
            help="additional context variables used in the template",
        )
        parser.add_argument(
            "--dry-run",
            default=False,
            action="store_true",
            help="Do not really save the CommunicationEventTypes",
        )
        parser.add_argument(
            "--show-html",
            default=False,
            action="store_true",
            help="Show html template also",
        )
        parser.add_argument(
            "--is-order-email",
            default=False,
            action="store_true",
            help="Use if the communicationevent is not user but order related",
        )

    def get_messages(self, code, options):
        ctx = {
            "user": DotPath("user"),
            "site": DotPath("site"),
            "logo": DotPath("logo"),
            "order": DotPath("order"),
            "status_url": DotPath("status_url"),
            "static_base_url": DotPath("static_base_url"),
            "tracking_url": DotPath("tracking_url"),
            "line": DotPath("line"),
            "field": DotPath("field"),
            "reset_url": DotPath("reset_url"),
            "new_email": DotPath("new_email"),
            "alert": DotPath("alert"),
            "lines": DotPath("lines"),
            "quantity": DotPath("quantity"),
            "option": DotPath("option"),
            "discount": DotPath("discount"),
            "charge": DotPath("charge"),
            "settings": DotPath("settings"),
            "request": DotPath("request"),
        }
        extra_context = options["variable"] or []

        for var in extra_context:
            ctx[var] = DotPath(var)

        # add EmailContext vaiables to context
        # this is casted back to dict, because if varables
        # in the ctx are DotPath, we don;t want some property
        # on EmailContext get priority, which could erase the
        # DotPath in ctx.
        email_ctx = dict(EmailContext(**ctx))
        messages = CommunicationEventType(code=code).get_messages(email_ctx)

        if options["is_order_email"]:
            messages["body"] = "{% load currency_filters %}\n" + messages["body"]
            if messages["html"]:
                messages["html"] = "{% load currency_filters %}\n" + messages["html"]
            else:
                messages["html"] = None

        return messages

    def handle(self, *args, **options):
        try:
            with transaction.atomic():
                try:
                    self._handle(**options)
                except ValueError:
                    # it failed because collectstatic made a booboo
                    call_command("collectstatic", "--noinput")
                    self._handle(**options)
        except (AttributeError, OperationalError):
            logger.warning("Unable to populate dashboard emails at this time")

    def _handle(self, **options):
        # make a copy of the tags for that can be changed and used in this command
        default_tags_copy = deepcopy(defaulttags.register.tags)
        wagtailimages_tags_copy = deepcopy(wagtailimages_tags.register.tags)
        wagtailsettings_tags_copy = deepcopy(wagtailsettings_tags.register.tags)
        ocyan_main_copy = deepcopy(ocyan_main.register.tags)

        with (
            mock.patch(
                "compressor.base.Compressor.template_name_inline",
                "override_compressor/css_inline.html",
                create=True,
            ),
            mock.patch.object(ocyan_main.register, "tags", ocyan_main_copy),
            mock.patch.object(
                wagtailsettings_tags.register, "tags", wagtailsettings_tags_copy
            ),
            mock.patch.object(
                wagtailimages_tags.register, "tags", wagtailimages_tags_copy
            ),
            mock.patch.object(defaulttags.register, "tags", default_tags_copy),
            mock.patch.object(template, "Parser", NoFilterParser),
        ):
            defaulttags.register.tag("load", execute_and_rerender(defaulttags.load))
            defaulttags.register.tag("if", do_not_change_anything)
            defaulttags.register.tag("else", do_not_change_anything)
            defaulttags.register.tag("elif", do_not_change_anything)
            defaulttags.register.tag("endif", do_not_change_anything)

            defaulttags.register.tag("for", do_not_change_anything)
            defaulttags.register.tag("endfor", do_not_change_anything)
            wagtailimages_tags.register.tag("image", do_not_change_anything)
            wagtailsettings_tags.register.tag("get_settings", do_not_change_anything)
            ocyan_main.register.tag("absolute_url", do_not_change_anything)

            with override(options["language_code"]):
                for code in options["event_types"]:
                    if (
                        CommunicationEventType.objects.filter(code=code).exists()
                        and not options["dry_run"]
                    ):
                        self.stdout.write(
                            self.style.WARNING("%s allready exists, skipping." % code)
                        )
                    else:
                        messages = self.get_messages(code, options)

                        if not messages["subject"] or not messages["body"]:
                            self.stdout.write(
                                self.style.ERROR(
                                    "skipping %s, no template found" % code
                                )
                            )
                            continue

                        if options["dry_run"]:
                            self.stdout.write(self.style.SUCCESS("subject:"))
                            self.stdout.write("%(subject)s\n\n" % messages)
                            self.stdout.write(self.style.SUCCESS("body:"))
                            self.stdout.write("%(body)s" % messages)
                            if options["show_html"]:
                                self.stdout.write(self.style.SUCCESS("html"))
                                self.stdout.write("%(html)s:" % messages)
                        else:
                            if options["is_order_email"]:
                                category = CommunicationEventType.ORDER_RELATED
                            else:
                                category = CommunicationEventType.USER_RELATED

                            evt = CommunicationEventType.objects.create(
                                code=code,
                                name=code,
                                category=category,
                                email_subject_template=messages["subject"],
                                email_body_template=messages["body"],
                                email_body_html_template=messages["html"],
                            )
                            if config.i18n_enabled:
                                for language_code, _ in settings.LANGUAGES:
                                    if language_code != options["language_code"]:
                                        with override(language_code):
                                            messages = self.get_messages(code, options)
                                            evt.email_subject_template = messages[
                                                "subject"
                                            ]
                                            evt.email_body_template = messages["body"]
                                            evt.email_body_html_template = messages[
                                                "html"
                                            ]
                                            evt.save()

        self.stdout.write(
            self.style.SUCCESS(
                "populating emails for events: %(event_types)s in language %(language_code)s done."
                % options
            )
        )



"""
The code in this file is there to ensure that the original migrations
of the app are loaded into the new migration module.

This ensures that you don't have to copy all migrations
when you modify your migration module to adjust/override a (abstract) model.

The migration loader uses pkgutil for migration dicovery.
That allows us to use pkgutil.extend_path to add the existing migrations.

A plugin extension can include migrations by registering a new migration module using settings.MIGRATION_MODULES,
you should use this to register the new migration module path containing the customizations.
It might be that this is allready in place by an extension or an override.
Please check if this is the case before writing you import by running `manage.py settings MIGRATION_MODULES`.
"""

import pkgutil
from oscar.apps.communication import migrations

__path__ = pkgutil.extend_path(__path__, migrations.__name__)



from oscar.apps.customer.views import OrderHistoryView as BaseOrderHistoryView

from ocyan.plugin.oscar.pipeline import PIPELINE_STATUS_CANCELLED, PIPELINE_STATUS_NEW


class OrderHistoryView(BaseOrderHistoryView):
    def get_queryset(self):
        qs = self.model._default_manager.filter(  # pylint: disable=W0212
            user=self.request.user
        ).exclude(status__in=[PIPELINE_STATUS_CANCELLED, PIPELINE_STATUS_NEW])
        if self.form.is_bound and self.form.is_valid():
            qs = qs.filter(**self.form.get_filters())
        return qs



from argparse import Namespace

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.sites.shortcuts import get_current_site
from django.views.generic import DetailView

from oscar.core.loading import get_model

from ocyan.plugin.oscar.info import OSCAR_NEEDS_OXYAN

from ocyan.plugin.oscar_communications.order_extension.context import (
    EmailContext,
    OrderContext,
    ShippingContext,
)

if OSCAR_NEEDS_OXYAN:
    CommunicationEventType = get_model("communication", "CommunicationEventType")
else:
    CommunicationEventType = get_model("customer", "CommunicationEventType")

Product = get_model("catalogue", "Product")
Order = get_model("order", "Order")
User = get_user_model()


class PreView(UserPassesTestMixin, DetailView):
    template_name = "communications/email-preview.html"

    def test_func(self):
        return self.request.user.is_superuser

    def get_template_names(self):
        return [
            "communications/email-preview.%(extension)s" % self.kwargs,
            self.template_name,
        ]

    def get_ctx(self):
        return EmailContext(**self.request.GET.dict())

    def get_context_data(self, **kwargs):
        event_type = self.kwargs["event_type"]
        messages = CommunicationEventType(code=event_type).get_messages(self.get_ctx())
        kwargs.update(messages)
        return super().get_context_data(**kwargs)


class OrderEmailPreview(PreView):
    model = Order
    template_name = "communications/email-preview.html"
    slug_field = "number"

    def get_ctx(self):
        order = self.get_object()

        extra_context = self.request.GET.dict()

        if order.shipping_events.exists():
            ctx = ShippingContext(order.shipping_events.first(), **extra_context)
        else:
            ctx = OrderContext(order, **extra_context)

        return ctx


class UserEmailPreview(PreView):
    model = User
    slug_field = "username"

    def get_ctx(self):
        user = self.get_object()
        alert = Namespace(product=Product.objects.first(), user=user)
        extra_context = self.request.GET.dict()

        context = {
            "new_email": user.email,
            "email": user.email,
            "static_base_url": getattr(settings, "OSCAR_STATIC_BASE_URL", None),
        }

        context.update(extra_context)

        return EmailContext(
            user=user, alert=alert, site=get_current_site(self.request), **context
        )



# pylint: disable=W0614,W0401
from oscar.core.loading import get_classes, is_model_registered

AbstractEmail, AbstractNotification = get_classes(
    "communication.abstract_models",
    ["AbstractEmail", "AbstractNotification"],
)

###############################################################################
# This is how to override an oscar app, while keeping the overriding app      #
# overridable itself.                                                         #
# The mechanism is the same as in oscar, the only think different is the way  #
# ``__all__`` is being updated with the exports of the original oscar app.    #
###############################################################################

__all__ = []

if not is_model_registered("communication", "Email"):

    class Email(AbstractEmail):
        pass

    __all__.append("Email")


if not is_model_registered("communication", "Notification"):

    class Notification(AbstractNotification):
        pass

    __all__.append("Notification")

# both these lines are reruired, the first one imports all objects the are not
# implemented above into the current namespace.
# the second line updates the __all__ variable of the current namespace with
# the objects from the oscar namespace.
# THIS ENSURE THAT THE SAME CAN BE DONE BY ANOTHER APP, SO THIS APP CAN BE ONCE
# MORE EXTENDED BY ANOTHER APP.
from oscar.apps.communication.models import *  # noqa
from oscar.apps.communication.models import __all__ as some

__all__.extend(some)



from django.urls import path

from .constants import APP_LABEL
from .views import OrderEmailPreview, UserEmailPreview

app_name = APP_LABEL
urlpatterns = [
    path(
        "order-email/<str:slug>/<str:event_type>/",
        OrderEmailPreview.as_view(),
        kwargs={"extension": "html"},
    ),
    path(
        "order-email/<str:slug>/<str:event_type>.<str:extension>",
        OrderEmailPreview.as_view(),
        name="order-email-detail",
    ),
    path(
        "user-email/<str:slug>/<str:event_type>/",
        UserEmailPreview.as_view(),
        kwargs={"extension": "html"},
    ),
    path(
        "user-email/<str:slug>/<str:event_type>.<str:extension>",
        UserEmailPreview.as_view(),
        name="user-email-detail",
    ),
]



import logging

from oscar.core.loading import get_class, get_model

from ocyan.plugin.oscar.info import OSCAR_NEEDS_OXYAN

logger = logging.getLogger(__name__)

if OSCAR_NEEDS_OXYAN:
    CommunicationEventType = get_model("communication", "CommunicationEventType")
else:
    CommunicationEventType = get_model("customer", "CommunicationEventType")

BccDispatcher = get_class("customer.utils", "BccDispatcher")


def send_message(ctx, code, bcc_admins=False, **kwargs):
    try:
        event_type = CommunicationEventType.objects.get(code=code)
    except CommunicationEventType.DoesNotExist:
        event_type = None

    if ctx.order.communication_events.filter(event_type=event_type).exists():
        logger.info("Not sending message for event %s, message has allready been sent.")
        return

    messages = CommunicationEventType.objects.get_and_render(code, ctx)

    if "attachments" in kwargs:
        messages["attachments"] = kwargs["attachments"]

    if messages and messages["body"]:
        logger.info("Order #%s - sending %s messages", ctx.order.number, code)
        dispatcher = BccDispatcher(bcc_admins, logger)
        dispatcher.dispatch_order_messages(ctx.order, messages, event_type, **kwargs)
    else:
        logger.warning(
            "Order #%s - no %s communication event type", ctx.order.number, code
        )



# -*- coding: utf-8 -*-
from django.apps import AppConfig
from django.core.management import call_command
from django.db.models.signals import post_migrate

from .constants import APP_LABEL

# if more communication templates are used, add them here
USER_COMMUNICATION_TYPES = [
    "PASSWORD_CHANGED",
    "EMAIL_CHANGED",
    "PASSWORD_RESET",
    "PRODUCT_ALERT",
    "PRODUCT_ALERT_CONFIRMATION",
    "REGISTRATION",
]

ORDER_COMMUNICATION_TYPES = ["ORDER_PLACED", "shipped"]


def populate_dashboard_emails(sender, app_config, **kwargs):  # pylint: disable=W0613
    if app_config.label == "customer":
        call_command("populate_dashboard_emails", *USER_COMMUNICATION_TYPES)
        call_command(
            "populate_dashboard_emails", "--is-order-email", *ORDER_COMMUNICATION_TYPES
        )


class OcyanPluginOscarCommunicationsAppConfig(AppConfig):
    name = "ocyan.plugin.oscar_communications"
    label = APP_LABEL

    def ready(self):
        post_migrate.connect(populate_dashboard_emails)
        # pylint: disable=unused-import
        from . import signal_handlers
